import dataclasses

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from scheduler.settings import conf_settings
from scheduler.tests.testtools import SchedulerBaseCase
from scheduler.types import Broker, SchedulerConfiguration


class TestWorkerAdmin(SchedulerBaseCase):
    def setUp(self):
        from scheduler.settings import SCHEDULER_CONFIG

        self.old_settings = SCHEDULER_CONFIG

    def tearDown(self):
        from scheduler import settings as scheduler_settings

        scheduler_settings.SCHEDULER_CONFIG = self.old_settings

    def test_scheduler_config_as_dict(self):
        from scheduler.settings import SCHEDULER_CONFIG

        settings.SCHEDULER_CONFIG = dict(
            EXECUTIONS_IN_PAGE=SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE + 1,
            SCHEDULER_INTERVAL=SCHEDULER_CONFIG.SCHEDULER_INTERVAL + 1,
            BROKER=Broker.REDIS,
            CALLBACK_TIMEOUT=SCHEDULER_CONFIG.SCHEDULER_INTERVAL + 1,
            DEFAULT_SUCCESS_TTL=SCHEDULER_CONFIG.DEFAULT_SUCCESS_TTL + 1,
            DEFAULT_FAILURE_TTL=SCHEDULER_CONFIG.DEFAULT_FAILURE_TTL + 1,
            DEFAULT_JOB_TTL=SCHEDULER_CONFIG.DEFAULT_JOB_TTL + 1,
            DEFAULT_JOB_TIMEOUT=SCHEDULER_CONFIG.DEFAULT_JOB_TIMEOUT + 1,
            # General configuration values
            DEFAULT_WORKER_TTL=SCHEDULER_CONFIG.DEFAULT_WORKER_TTL + 1,
            DEFAULT_MAINTENANCE_TASK_INTERVAL=SCHEDULER_CONFIG.DEFAULT_MAINTENANCE_TASK_INTERVAL + 1,
            DEFAULT_JOB_MONITORING_INTERVAL=SCHEDULER_CONFIG.DEFAULT_JOB_MONITORING_INTERVAL + 1,
            SCHEDULER_FALLBACK_PERIOD_SECS=SCHEDULER_CONFIG.SCHEDULER_FALLBACK_PERIOD_SECS + 1,
        )
        conf_settings()
        from scheduler.settings import SCHEDULER_CONFIG

        for key, value in settings.SCHEDULER_CONFIG.items():
            self.assertEqual(getattr(SCHEDULER_CONFIG, key), value)

    def test_scheduler_config_as_data_class(self):
        from scheduler.settings import SCHEDULER_CONFIG

        self.assertEqual(SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE, 20)
        settings.SCHEDULER_CONFIG = SchedulerConfiguration(
            EXECUTIONS_IN_PAGE=1,
            SCHEDULER_INTERVAL=60,
            BROKER=Broker.REDIS,
            CALLBACK_TIMEOUT=1111,
            DEFAULT_SUCCESS_TTL=1111,
            DEFAULT_FAILURE_TTL=111111,
            DEFAULT_JOB_TTL=1111,
            DEFAULT_JOB_TIMEOUT=11111,
            # General configuration values
            DEFAULT_WORKER_TTL=11111,
            DEFAULT_MAINTENANCE_TASK_INTERVAL=111,
            DEFAULT_JOB_MONITORING_INTERVAL=1111,
            SCHEDULER_FALLBACK_PERIOD_SECS=1111,
        )
        conf_settings()
        from scheduler.settings import SCHEDULER_CONFIG

        for key, value in dataclasses.asdict(settings.SCHEDULER_CONFIG).items():
            self.assertEqual(getattr(SCHEDULER_CONFIG, key), value)

    def test_scheduler_config_as_dict_bad_param(self):
        settings.SCHEDULER_CONFIG = dict(
            EXECUTIONS_IN_PAGE=1,
            SCHEDULER_INTERVAL=60,
            BROKER=Broker.REDIS,
            CALLBACK_TIMEOUT=1111,
            DEFAULT_SUCCESS_TTL=1111,
            DEFAULT_FAILURE_TTL=111111,
            DEFAULT_JOB_TTL=1111,
            DEFAULT_JOB_TIMEOUT=11111,
            # General configuration values
            DEFAULT_WORKER_TTL=11111,
            DEFAULT_MAINTENANCE_TASK_INTERVAL=111,
            DEFAULT_JOB_MONITORING_INTERVAL=1111,
            SCHEDULER_FALLBACK_PERIOD_SECS=1111,
            BAD_PARAM="bad_value",  # This should raise an error
        )
        self.assertRaises(ImproperlyConfigured, conf_settings)



from datetime import datetime
from typing import Callable, Any, Tuple, Dict, Type

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from scheduler.helpers import utils

ARG_TYPE_TYPES_DICT: Dict[str, Type] = {
    "str": str,
    "int": int,
    "bool": bool,
    "datetime": datetime,
    "callable": Callable,
}


class BaseTaskArg(models.Model):
    class ArgType(models.TextChoices):
        STR = "str", _("string")
        INT = "int", _("int")
        BOOL = "bool", _("boolean")
        DATETIME = "datetime", _("datetime")
        CALLABLE = "callable", _("callable")

    arg_type = models.CharField(
        _("Argument Type"),
        max_length=12,
        choices=ArgType.choices,
        default=ArgType.STR,
    )
    val = models.CharField(_("Argument Value"), blank=True, max_length=255)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey()

    def clean(self) -> None:
        if self.arg_type not in ARG_TYPE_TYPES_DICT:
            raise ValidationError(
                {
                    "arg_type": ValidationError(
                        _(f"Could not parse {self.arg_type}, options are: {ARG_TYPE_TYPES_DICT.keys()}"), code="invalid"
                    )
                }
            )
        try:
            if self.arg_type == "callable":
                utils.callable_func(self.val)
            elif self.arg_type == "datetime":
                datetime.fromisoformat(self.val)
            elif self.arg_type == "bool":
                if self.val.lower() not in {"true", "false"}:
                    raise ValidationError
            elif self.arg_type == "int":
                int(self.val)
        except Exception:
            raise ValidationError(
                {"arg_type": ValidationError(_(f"Could not parse {self.val} as {self.arg_type}"), code="invalid")}
            )

    def save(self, **kwargs: Any) -> None:
        super(BaseTaskArg, self).save(**kwargs)
        self.content_object.save()

    def delete(self, **kwargs: Any) -> None:
        super(BaseTaskArg, self).delete(**kwargs)
        self.content_object.save()

    def value(self) -> Any:
        if self.arg_type == "callable":
            res = utils.callable_func(self.val)()
        elif self.arg_type == "datetime":
            res = datetime.fromisoformat(self.val)
        elif self.arg_type == "bool":
            res = self.val.lower() == "true"
        else:
            res = ARG_TYPE_TYPES_DICT[self.arg_type](self.val)
        return res

    class Meta:
        abstract = True
        ordering = ["id"]


class TaskArg(BaseTaskArg):
    def __str__(self) -> str:
        return f"TaskArg[arg_type={self.arg_type},value={self.value()}]"


class TaskKwarg(BaseTaskArg):
    key = models.CharField(max_length=255)

    def __str__(self) -> str:
        key, value = self.value()
        return f"TaskKwarg[key={key},arg_type={self.arg_type},value={self.val}]"

    def value(self) -> Tuple[str, Any]:
        return self.key, super(TaskKwarg, self).value()



import logging
import os
import sys
from argparse import ArgumentParser
from typing import Any, Optional

import click
from django.core.management.base import BaseCommand
from django.db import connections

from scheduler.settings import logger
from scheduler.types import ConnectionErrorTypes
from scheduler.worker import create_worker

VERBOSITY_TO_LOG_LEVEL = {
    0: logging.CRITICAL,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}

WORKER_ARGUMENTS = {
    "queues",
    "name",
    "connection",
    "maintenance_interval",
    "job_monitoring_interval",
    "dequeue_strategy",
    "disable_default_exception_handler",
    "fork_job_execution",
    "with_scheduler",
    "burst",
}


def reset_db_connections() -> None:
    for c in connections.all():
        c.close()


def register_sentry(sentry_dsn: str, **opts: Any) -> None:
    try:
        import sentry_sdk
        from scheduler.helpers.sentry_integration import SentryIntegration
    except ImportError:
        logger.error("Sentry SDK not installed. Skipping Sentry Integration")
        return

    sentry_sdk.init(sentry_dsn, integrations=[SentryIntegration()], **opts)


class Command(BaseCommand):
    """Runs scheduler workers on specified queues.
    Note that all queues passed into a single scheduler_worker command must share the same connection.

    Example usage:
    python manage.py scheduler_worker high medium low
    """

    args = "<queue queue ...>"

    def _add_sentry_args(self, parser: ArgumentParser) -> None:
        parser.add_argument("--sentry-dsn", action="store", dest="sentry_dsn", help="Sentry DSN to use")
        parser.add_argument("--sentry-debug", action="store_true", dest="sentry_debug", help="Enable Sentry debug mode")
        parser.add_argument("--sentry-ca-certs", action="store", dest="sentry_ca_certs", help="Path to CA certs file")

    def _add_work_args(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--burst", action="store_true", dest="burst", default=False, help="Run worker in burst mode"
        )
        parser.add_argument(
            "--max-jobs",
            action="store",
            default=None,
            dest="max_jobs",
            type=int,
            help="Maximum number of jobs to execute before terminating worker",
        )
        parser.add_argument(
            "--max-idle-time",
            action="store",
            default=None,
            dest="max_idle_time",
            type=int,
            help="Maximum number of seconds to wait for new job before terminating worker",
        )
        parser.add_argument(
            "--without-scheduler",
            action="store_false",
            default=True,
            dest="with_scheduler",
            help="Run worker without scheduler, default to with scheduler",
        )

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--pid", action="store", dest="pidfile", default=None, help="file to write the worker`s pid into"
        )
        parser.add_argument("--name", action="store", dest="name", default=None, help="Name of the worker")
        parser.add_argument(
            "--worker-ttl",
            action="store",
            type=int,
            dest="worker_ttl",
            default=420,
            help="Default worker timeout to be used",
        )
        parser.add_argument(
            "--fork-job-execution",
            action="store",
            default=True,
            dest="fork_job_execution",
            type=bool,
            help="Fork job execution to another process",
        )
        parser.add_argument(
            "queues",
            nargs="*",
            type=str,
            help="The queues to work on, separated by space, all queues should be using the same redis",
        )
        self._add_sentry_args(parser)
        self._add_work_args(parser)

    def handle(self, **options: Any) -> None:
        queues = options.pop("queues", [])
        if not queues:
            queues = [
                "default",
            ]
        click.echo(f"Starting worker for queues {queues}")
        pidfile = options.pop("pidfile")
        if pidfile:
            with open(os.path.expanduser(pidfile), "w") as fp:
                fp.write(str(os.getpid()))

        # Verbosity is defined by default in BaseCommand for all commands
        verbosity = options.pop("verbosity", 3)
        log_level = VERBOSITY_TO_LOG_LEVEL.get(verbosity, logging.INFO)
        logger.setLevel(log_level)

        init_options = {k: v for k, v in options.items() if k in WORKER_ARGUMENTS}

        try:
            # Instantiate a worker
            w = create_worker(*queues, **init_options)

            # Close any opened DB connection before any fork
            reset_db_connections()

            # Check whether sentry is enabled
            if options.get("sentry_dsn") is not None:
                sentry_opts = dict(ca_certs=options.get("sentry_ca_certs"), debug=options.get("sentry_debug"))
                dsn: str = options.get("sentry_dsn")  # type: ignore
                register_sentry(dsn, **sentry_opts)

            max_jobs: Optional[int] = options.get("max_jobs", None)
            max_idle_time: Optional[int] = options.get("max_idle_time", None)
            w.work(max_jobs=max_jobs, max_idle_time=max_idle_time)
        except ConnectionErrorTypes as e:
            click.echo(str(e), err=True)
            sys.exit(1)



from datetime import datetime
from typing import List, Union

from django.contrib import admin, messages
from django.contrib.contenttypes.admin import GenericStackedInline
from django.db.models import QuerySet
from django.http import HttpRequest, HttpResponse
from django.utils import timezone, formats
from django.utils.timezone import is_naive
from django.utils.translation import gettext_lazy as _

from scheduler.helpers.queues import get_queue
from scheduler.models import TaskArg, TaskKwarg, Task, TaskType
from scheduler.redis_models import JobModel
from scheduler.settings import SCHEDULER_CONFIG, logger
from scheduler.types import ConnectionErrorTypes


def job_execution_of(job: JobModel, task: Task) -> bool:
    return job.scheduled_task_id == task.id and job.task_type == task.task_type


def get_job_executions_for_task(queue_name: str, scheduled_task: Task) -> List[JobModel]:
    queue = get_queue(queue_name)
    job_list: List[JobModel] = list(
        filter(lambda job: job is not None, JobModel.get_many(queue.get_all_job_names(), connection=queue.connection))
    )

    res = sorted(
        list(filter(lambda j: job_execution_of(j, scheduled_task), job_list)), key=lambda j: j.created_at, reverse=True
    )
    return res


class JobArgInline(GenericStackedInline):
    model = TaskArg
    extra = 0
    fieldsets = ((None, dict(fields=("arg_type", "val"))),)


class JobKwargInline(GenericStackedInline):
    model = TaskKwarg
    extra = 0
    fieldsets = ((None, dict(fields=("key", ("arg_type", "val")))),)


def get_message_bit(rows_updated: int) -> str:
    message_bit = "1 task was" if rows_updated == 1 else f"{rows_updated} tasks were"
    return message_bit


@admin.register(Task)
class TaskAdmin(admin.ModelAdmin):
    """TaskAdmin admin view for all task models."""

    class Media:
        js = ("admin/js/jquery.init.js", "admin/js/select-fields.js")

    save_on_top = True
    change_form_template = "admin/scheduler/change_form.html"
    actions = ["disable_selected", "enable_selected", "enqueue_job_now"]
    inlines = [JobArgInline, JobKwargInline]
    list_filter = ("enabled", "task_type", "queue")
    list_display = (
        "enabled",
        "name",
        "function_string",
        "is_scheduled",
        "queue",
        "task_schedule",
        "next_run",
        "successful_runs",
        "last_successful_run",
        "failed_runs",
        "last_failed_run",
    )
    list_display_links = ("name",)
    readonly_fields = (
        "job_name",
        "successful_runs",
        "last_successful_run",
        "failed_runs",
        "last_failed_run",
    )
    # radio_fields = {"task_type": admin.HORIZONTAL}
    fieldsets = (
        (
            None,
            dict(
                fields=(
                    "name",
                    "callable",
                    ("enabled", "timeout", "result_ttl"),
                    "task_type",
                )
            ),
        ),
        (
            None,
            dict(fields=("scheduled_time",), classes=("tasktype-OnceTaskType",)),
        ),
        (
            None,
            dict(fields=("cron_string",), classes=("tasktype-CronTaskType",)),
        ),
        (
            None,
            dict(
                fields=(
                    (
                        "interval",
                        "interval_unit",
                    ),
                    "repeat",
                ),
                classes=("tasktype-RepeatableTaskType",),
            ),
        ),
        (_("Queue settings"), dict(fields=(("queue", "at_front"), "job_name"))),
        (
            _("Previous runs info"),
            dict(fields=(("successful_runs", "last_successful_run"), ("failed_runs", "last_failed_run"))),
        ),
    )

    @admin.display(description="Schedule")
    def task_schedule(self, o: Task) -> str:
        if o.task_type == TaskType.ONCE.value:
            if timezone.is_naive(o.scheduled_time):
                local_time = timezone.make_aware(o.scheduled_time, timezone.get_current_timezone())
            else:
                local_time = timezone.localtime(o.scheduled_time)
            return f"Run once: {formats.date_format(local_time, 'DATETIME_FORMAT')}"
        elif o.task_type == TaskType.CRON.value:
            return f"Cron: {o.cron_string}"
        else:  # if o.task_type == TaskType.REPEATABLE.value:
            if o.interval is None or o.interval_unit is None:
                return ""
            return f"Repeatable: {o.interval} {o.get_interval_unit_display()}"

    @admin.display(description="Next run")
    def next_run(self, o: Task) -> Union[str, datetime]:
        res = o.scheduled_time
        if res < timezone.now():
            o.save(clean=False)
            res = o.scheduled_time
        if res is None:
            return _("Not scheduled")
        if is_naive(res):
            res = timezone.make_aware(res, timezone.get_current_timezone())
        return res

    def change_view(self, request: HttpRequest, object_id, form_url="", extra_context=None) -> HttpResponse:
        extra = extra_context or {}
        obj = self.get_object(request, object_id)
        try:
            execution_list = get_job_executions_for_task(obj.queue, obj)
        except ConnectionErrorTypes as e:
            logger.warn(f"Could not get job executions: {e}")
            execution_list = list()
        paginator = self.get_paginator(request, execution_list, SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE)
        page_number = request.GET.get("p", 1)
        page_obj = paginator.get_page(page_number)
        page_range = paginator.get_elided_page_range(page_obj.number)

        extra.update(
            {
                "pagination_required": paginator.count > SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE,
                "executions": page_obj,
                "page_range": page_range,
                "page_var": "p",
            }
        )

        return super(TaskAdmin, self).change_view(request, object_id, form_url, extra_context=extra)

    def delete_queryset(self, request: HttpRequest, queryset: QuerySet) -> None:
        for job in queryset:
            job.unschedule()
        super(TaskAdmin, self).delete_queryset(request, queryset)

    def delete_model(self, request: HttpRequest, obj: Task) -> None:
        obj.unschedule()
        super(TaskAdmin, self).delete_model(request, obj)

    @admin.action(description=_("Disable selected %(verbose_name_plural)s"), permissions=("change",))
    def disable_selected(self, request: HttpRequest, queryset: QuerySet) -> None:
        rows_updated = 0
        for obj in queryset.filter(enabled=True).iterator():
            obj.enabled = False
            obj.unschedule()
            rows_updated += 1

        level = messages.WARNING if not rows_updated else messages.INFO
        self.message_user(
            request, f"{get_message_bit(rows_updated)} successfully disabled and unscheduled.", level=level
        )

    @admin.action(description=_("Enable selected %(verbose_name_plural)s"), permissions=("change",))
    def enable_selected(self, request: HttpRequest, queryset: QuerySet) -> None:
        rows_updated = 0
        for obj in queryset.filter(enabled=False).iterator():
            obj.enabled = True
            obj.save()
            rows_updated += 1

        level = messages.WARNING if not rows_updated else messages.INFO
        self.message_user(request, f"{get_message_bit(rows_updated)} successfully enabled and scheduled.", level=level)

    @admin.action(description="Enqueue now", permissions=("change",))
    def enqueue_job_now(self, request: HttpRequest, queryset: QuerySet) -> None:
        task_names = []
        for task in queryset:
            task.enqueue_to_run()
            task_names.append(task.name)
        self.message_user(
            request,
            f"The following jobs have been enqueued: {', '.join(task_names)}",
        )



import weakref
from typing import Any, Callable

import sentry_sdk
from sentry_sdk._types import EventProcessor, Event, ExcInfo
from sentry_sdk.api import continue_trace
from sentry_sdk.consts import OP
from sentry_sdk.integrations import _check_minimum_version, Integration
from sentry_sdk.integrations.logging import ignore_logger
from sentry_sdk.tracing import TransactionSource
from sentry_sdk.utils import (
    capture_internal_exceptions,
    ensure_integration_enabled,
    event_from_exception,
    parse_version,
)

import scheduler
from scheduler.helpers.queues import Queue
from scheduler.redis_models import JobModel
from scheduler.redis_models import JobStatus
from scheduler.timeouts import JobTimeoutException
from scheduler.worker import Worker


class SentryIntegration(Integration):
    identifier = "rq"
    origin = f"auto.queue.{identifier}"

    @staticmethod
    def setup_once() -> None:
        version = parse_version(scheduler.__version__)
        _check_minimum_version(SentryIntegration, version)

        old_perform_job = Worker.perform_job

        @ensure_integration_enabled(SentryIntegration, old_perform_job)
        def sentry_patched_perform_job(self: Any, job_model: JobModel, *args: Queue, **kwargs: Any) -> bool:
            with sentry_sdk.new_scope() as scope:
                scope.clear_breadcrumbs()
                scope.add_event_processor(_make_event_processor(weakref.ref(job_model)))

                transaction = continue_trace(
                    job_model.meta.get("_sentry_trace_headers") or {},
                    op=OP.QUEUE_TASK_RQ,
                    name="unknown RQ task",
                    source=TransactionSource.TASK,
                    origin=SentryIntegration.origin,
                )

                with capture_internal_exceptions():
                    transaction.name = job_model.func_name

                with sentry_sdk.start_transaction(
                    transaction,
                    custom_sampling_context={"rq_job": job_model},
                ):
                    rv = old_perform_job(self, job_model, *args, **kwargs)

            if self.is_horse:
                # We're inside of a forked process and RQ is
                # about to call `os._exit`. Make sure that our
                # events get sent out.
                sentry_sdk.get_client().flush()

            return rv

        Worker.perform_job = sentry_patched_perform_job  # type: ignore[method-assign]

        old_handle_exception = Worker.handle_exception

        def sentry_patched_handle_exception(self: Worker, job: Any, *exc_info: Any, **kwargs: Any) -> Any:
            retry = hasattr(job, "retries_left") and job.retries_left and job.retries_left > 0
            failed = job._status == JobStatus.FAILED or job.is_failed
            if failed and not retry:
                _capture_exception(exc_info)

            return old_handle_exception(self, job, *exc_info, **kwargs)

        Worker.handle_exception = sentry_patched_handle_exception  # type: ignore[method-assign]

        old_enqueue_job = Queue.enqueue_job

        @ensure_integration_enabled(SentryIntegration, old_enqueue_job)
        def sentry_patched_enqueue_job(self: Queue, job: Any, **kwargs: Any) -> Any:
            scope = sentry_sdk.get_current_scope()
            if scope.span is not None:
                job.meta["_sentry_trace_headers"] = dict(scope.iter_trace_propagation_headers())

            return old_enqueue_job(self, job, **kwargs)

        Queue.enqueue_job = sentry_patched_enqueue_job  # type: ignore[method-assign]

        ignore_logger("rq.worker")


def _make_event_processor(weak_job: Callable[[], JobModel]) -> EventProcessor:
    def event_processor(event: Event, hint: dict[str, Any]) -> Event:
        job = weak_job()
        if job is not None:
            with capture_internal_exceptions():
                extra = event.setdefault("extra", {})
                extra["job"] = job.serialize()

        if "exc_info" in hint:
            with capture_internal_exceptions():
                if issubclass(hint["exc_info"][0], JobTimeoutException):
                    event["fingerprint"] = ["django-tasks-scheduler", "JobTimeoutException", job.func_name]

        return event

    return event_processor


def _capture_exception(exc_info: ExcInfo, **kwargs: Any) -> None:
    client = sentry_sdk.get_client()

    event, hint = event_from_exception(
        exc_info,
        client_options=client.options,
        mechanism={"type": "rq", "handled": False},
    )

    sentry_sdk.capture_event(event, hint=hint)



import datetime
import importlib
import time
from typing import Callable, Any


def current_timestamp() -> int:
    """Returns current UTC timestamp in secs"""
    return int(time.time())


def utcnow() -> datetime.datetime:
    """Return now in UTC"""
    return datetime.datetime.now(datetime.timezone.utc)


def callable_func(callable_str: str) -> Callable[[Any], Any]:
    path = callable_str.split(".")
    module = importlib.import_module(".".join(path[:-1]))
    func: Callable[[Any], Any] = getattr(module, path[-1])
    if not callable(func):
        raise TypeError(f"'{callable_str}' is not callable")
    return func



from datetime import timedelta, datetime

import bs4
import time_machine
from django.core.exceptions import ValidationError
from django.urls import reverse
from django.utils import timezone

from scheduler import settings
from scheduler.helpers.queues import get_queue
from scheduler.models import TaskType
from scheduler.redis_models import JobModel
from scheduler.tests.test_task_types.test_task_model import BaseTestCases
from scheduler.tests.testtools import task_factory


class TestScheduledOnceTask(BaseTestCases.TestSchedulableTask):
    task_type = TaskType.ONCE
    queue_name = settings.get_queue_names()[0]

    def test_clean(self):
        task = task_factory(self.task_type)
        task.queue = self.queue_name
        task.callable = "scheduler.tests.jobs.test_job"
        self.assertIsNone(task.clean())

    @time_machine.travel(datetime(2016, 12, 25))
    def test_admin_changelist_view__has_timezone_data(self):
        # arrange
        self.client.login(username="admin", password="admin")
        task_factory(self.task_type)
        url = reverse("admin:scheduler_task_changelist")
        # act
        res = self.client.get(url)
        # assert
        self.assertContains(res, "Run once: Dec. 26, 2016, midnight", count=1, status_code=200)

    @time_machine.travel(datetime(2016, 12, 25))
    def test_admin_change_view__has_execution_list(self):
        # arrange
        self.client.login(username="admin", password="admin")
        task = task_factory(self.task_type)
        url = reverse("admin:scheduler_task_change", args=(task.id,))
        # act
        res = self.client.get(url)
        # assert
        self.assertEqual(200, res.status_code)
        self.assertContains(res, "Job executions")
        self.assertFalse(res.context["pagination_required"])
        self.assertEqual(res.context["executions"].paginator.count, 1)
        soup = bs4.BeautifulSoup(res.content, "html.parser")
        self.assertEqual(1, len(soup.find_all("table", {"id": "result_list"})))
        counter_element_list = soup.find_all("span", {"id": "counter"})
        self.assertEqual(1, len(counter_element_list))
        counter_element = counter_element_list[0]
        self.assertEqual("1 entry", counter_element.text.strip())

    @time_machine.travel(datetime(2016, 12, 25))
    def test_admin_change_view__has_empty_execution_list(self):
        # arrange
        self.client.login(username="admin", password="admin")
        task = task_factory(self.task_type)
        queue = get_queue(self.queue_name)
        job = JobModel.get(task.job_name, connection=queue.connection)
        JobModel.delete(job, connection=queue.connection)
        url = reverse("admin:scheduler_task_change", args=(task.id,))
        # act
        res = self.client.get(url)
        # assert
        self.assertContains(res, "Job executions")
        self.assertContains(res, """<table id="result_list">""")
        self.assertContains(res, """<span id="counter">0""", status_code=200)
        self.assertFalse(res.context["pagination_required"])
        self.assertEqual(res.context["executions"].paginator.count, 0)

    def test_create_without_date__fail(self):
        task = task_factory(self.task_type, scheduled_time=None, instance_only=True)
        self.assertIsNone(task.scheduled_time)
        with self.assertRaises(Exception) as cm:
            task.clean()
        self.assertTrue(isinstance(cm.exception, ValidationError))
        self.assertEqual(str(cm.exception), "{'scheduled_time': ['Scheduled time is required']}")

    def test_create_with_date_in_the_past__fail(self):
        task = task_factory(self.task_type, scheduled_time=datetime.now() - timedelta(days=1), instance_only=True)
        with self.assertRaises(Exception) as cm:
            task.clean()
        self.assertTrue(isinstance(cm.exception, ValidationError))
        self.assertEqual(str(cm.exception), "{'scheduled_time': ['Scheduled time must be in the future']}")

    def test_unschedulable_old_job(self):
        task = task_factory(self.task_type, scheduled_time=timezone.now() - timedelta(hours=1), instance_only=True)
        task.save(clean=False)
        self.assertFalse(task.is_scheduled())
